#ifndef _DWC_IPC_C
#define _DWC_IPC_C

#ifndef lint
static const char rcsid[] = "$Id: dwc_ipc.c,v 1.12 2001/06/19 00:11:31 cvser Exp $";
#endif

#include <errno.h>
#include <string.h>

#include "dwc_lib.h"
#include "dwc_err.h"
#include "dwc_macro.h"
#include "dwc_timo.h"
#include "dwc_ipc.h"
#include "dwc_tcp.h"
#if defined(__UNIX_DW) 
#include "dwc_unix.h"
#endif /* __UNIX_DW */

#if defined(__WIN32_DW)
#define ETIMEDOUT   EIO
#endif /* __WIN32_DW */

typedef union {
    DWC_tcp_socket_t        hnd_tcp;
#if defined(__UNIX_DW) 
    DWC_unix_socket_t       hnd_unix;
#endif
} DWC_ipc_handle_t;

typedef union {
    DWC_tcp_socket_addr_t   sa_tcp;
#if defined(__UNIX_DW) 
    DWC_unix_socket_addr_t  sa_unix;
#endif
} DWC_ipc_address_t;

#define DW_MAX_ADDR_LEN     (256)

typedef struct {
    DWC_char_t              name[DW_MAX_ADDR_LEN];
    DWC_length_t            name_len;
    DWC_ipc_address_t       addr;
} DWC_ipc_addr_t;

#define NULL_IPC_ADDR( a_ipc )\
BEGIN_MACRO\
    (a_ipc).name[0] = NULLCHAR;\
    (a_ipc).name_len = 0;\
    memset( &(a_ipc).addr, 0, sizeof(DWC_ipc_address_t) );\
END_MACRO

typedef DWC_error_code_t   (*DWC_SendFunc_t)();
typedef DWC_error_code_t   (*DWC_RecvFunc_t)();
typedef DWC_error_code_t   (*DWC_WaitFunc_t)();
typedef DWC_string_t       (*DWC_LastErrFunc_t)(void);

struct _DWC_ipc {
    DWC_ipc_type_t          type;
    DWC_bool_t              allocated;
    DWC_bool_t              connected;
    DWC_bool_t              nonblocking;
    DWC_bool_t              listed;
    DWC_bool_t              binded;
    DWC_ipc_handle_t        handle;
    DWC_ipc_addr_t          local;
    DWC_ipc_addr_t          remote;
    DWC_SendFunc_t          send_func;
    DWC_RecvFunc_t          recv_func;
    DWC_WaitFunc_t          wait_func;
    DWC_TimeoutProc_t       cancel_call;
    DWC_LastErrFunc_t       error_func;
};

typedef struct {
    unsigned int            tcp_sock:1,
        unix_sock:1;
} DWC_ipc_subsys_mode_t;

typedef struct {
    unsigned int            inited:1;
    int                     error;
    DWC_bool_t              errored;
} DWC_ipc_subsys_state_t;

static DWC_ipc_subsys_mode_t  ipcMode = { 0 };
static DWC_ipc_subsys_state_t ipcState = { 0 };

/* -------------------------------------------------- */
DWC_error_code_t init_ipc_subsystem ( void )
{
    int ret;
    
    if( ipcState.inited == ENABLE )
    {
        dprintLib( "IPC system already inited.\n" );
        return DWC_ALREADYPERFORM;
    }
    ret = InitTimerEngine();
    if( ret != DWC_NOERROR )
    {
        dprintLib( "IPC: can't init timers engine.\n" );
        return DWC_IPCSYSTEM;
    }
    ret = init_tcp_sockets();
    if( ret == DWC_NOERROR )
    {
        ipcMode.tcp_sock = ENABLE;
    }
    else
    {
        dprintLib( "Can't initialize TCP subsystem (%d -- %s).\n", ret, last_tcp_error() );
        ipcMode.tcp_sock = DISABLE;
    }
#if defined(__UNIX_DW) 
    ret = init_unix_sockets();
    if( ret == DWC_NOERROR )
    {
        ipcMode.unix_sock = ENABLE;
    }
    else
    {
        dprintLib( "Can't initialize UNIX-SOCKET subsystem (%d -- %s).\n", ret, last_unix_error() );
        ipcMode.unix_sock = DISABLE;
    }
#else    
    ipcMode.unix_sock = DISABLE;
#endif
    ipcState.inited = ENABLE;
    return DWC_NOERROR;
}

/* -------------------------------------------------- */
DWC_error_code_t dinit_ipc_subsystem ( void )
{
    int tret = DWC_NOERROR;
    int uret = DWC_NOERROR;
    int eret = DWC_NOERROR;

    if( ipcState.inited != ENABLE )
    {
        dprintLib( "IPC system wasn't inited.\n" );
        return DWC_NOTREADY;
    }
    eret = DeInitTimerEngine();
    if( eret != DWC_NOERROR )
    {
        dprintLib( "IPC: can't deinit timers engine.\n" );
    }
    if( ipcMode.tcp_sock == ENABLE )
    {
        tret = dinit_tcp_sockets();
    }
    else
    {
        tret = DWC_NOERROR;
    }
    if( tret != DWC_NOERROR )
    {
        dprintLib( "Can't normally deinitialize TCP subsystem ( %d -- %s ).\n", tret, last_tcp_error() );
    }
#if defined(__UNIX_DW)
    if( ipcMode.unix_sock == DISABLE )
    {
        uret = dinit_unix_sockets();
    }
    else
    {
        uret = DWC_NOERROR;
    }
    if( uret != DWC_NOERROR )
    {
        dprintLib( "Can't normally deinitialize UNIX-SOCKETS subsystem ( %d -- %s ).\n", uret, last_unix_error() );
    }
#else
	uret = DWC_NOERROR;
#endif
    return ( eret == DWC_NOERROR ? ( tret == DWC_NOERROR ? uret : tret ) : eret );
}

/* -------------------------------------------------- */
DWC_error_code_t create_ipc ( DWC_ipc_type_t type, DWC_ipc_t *ipc )
{
    DWC_error_code_t error;
    
    PARAM_ASSERT( ipc );
    PARAM_ASSERT( ((DWC_IPC_TYPE_0 < type) && (type < DWC_IPC_TYPE_MAX)) );

    if( !(*ipc) )
    {
        (*ipc) = DWC_Malloc( DWC_ipc_t, sizeof(struct _DWC_ipc) );
        if( !ipc )
        {
            return DWC_NOMEMORY;
        }
        (*ipc)->allocated = TRUE;
    }
    else
    {
        (*ipc)->allocated = FALSE;
    }
    (*ipc)->type = type;
    (*ipc)->connected = FALSE;
    (*ipc)->binded = FALSE;
#if defined(__WIN32_DW)
    (*ipc)->nonblocking = TRUE;
#else /* __WIN32_DW */
    (*ipc)->nonblocking = FALSE;
#endif /* __WIN32_DW */
    (*ipc)->listed = FALSE;
    NULL_IPC_ADDR( (*ipc)->local );
    NULL_IPC_ADDR( (*ipc)->remote );
    switch( type )
    {
        case DWC_IPC_TCP_SOCKET:
            if( ipcMode.tcp_sock != ENABLE )
            {
                error = DWC_BADIPC;
                goto error_exit;
            }
            error = create_tcp_socket( &(*ipc)->handle.hnd_tcp, (*ipc)->nonblocking );
            if( error != DWC_NOERROR )
            {
                goto error_exit;
            }
            (*ipc)->send_func = send_tcp_socket;
            (*ipc)->recv_func = recv_tcp_socket;            
            (*ipc)->wait_func = wait_tcp_socket;            
            (*ipc)->cancel_call = cancel_tcp_call;
            (*ipc)->error_func = last_tcp_error;
            (*ipc)->local.addr.sa_tcp.sin_family = PF_INET;
            (*ipc)->remote.addr.sa_tcp.sin_family = PF_INET;
            break;
#if defined(__UNIX_DW)            
        case DWC_IPC_UNIX_SOCKET:
            if( ipcMode.unix_sock != ENABLE )
            {
                error = DWC_BADIPC;
                goto error_exit;
            }
            error = create_unix_socket( &(*ipc)->handle.hnd_unix, (*ipc)->nonblocking );
            if( error != DWC_NOERROR )
            {
                goto error_exit;
            }
            (*ipc)->send_func = send_unix_socket;
            (*ipc)->recv_func = recv_unix_socket;
            (*ipc)->wait_func = wait_unix_socket;            
            (*ipc)->cancel_call = cancel_unix_call;
            (*ipc)->error_func = last_unix_error;
            (*ipc)->local.addr.sa_unix.sun_family = PF_UNIX;
            (*ipc)->remote.addr.sa_unix.sun_family = PF_UNIX;
            break;
#endif /* __UNIX_DW */
        default:
            return DWC_BADIPC; 
    }    
    return DWC_NOERROR;

  error_exit:
    delete_ipc( ipc );
    return error;
}

/* -------------------------------------------------- */
DWC_error_code_t delete_ipc ( DWC_ipc_t* ipc )
{
    DWC_error_code_t error;

    PARAM_ASSERT( ipc );
    
    if( !(*ipc) )
        return DWC_BADIPC;
    switch( (*ipc)->type )
    {
        case DWC_IPC_TCP_SOCKET:
            error = close_tcp_socket( &(*ipc)->handle.hnd_tcp );
            break;
#if defined(__UNIX_DW)
        case DWC_IPC_UNIX_SOCKET:
            error = close_unix_socket( &(*ipc)->handle.hnd_unix, (*ipc)->local.addr.sa_unix, (*ipc)->binded );
            break;
#endif /* __UNIX_DW */
        default:
            return DWC_BADIPC;
    }
    if( (*ipc)->allocated )
        DWC_Free( *ipc );
    return DWC_NOERROR;
}

/* -------------------------------------------------- */
DWC_error_code_t addr_ipc_op ( DWC_ipc_t ipc, DWC_cmd_t cmd, DWC_side_t side, DWC_string_t name, DWC_length_t name_len, ... )
{    
    DWC_ipc_addr_t      *fulladdr;
    DWC_error_code_t     error;    
    va_list              args;
    
    if( (!name && (name_len != 0)) 
        || 
        (name && (name_len == 0))
        ||
        (name && (name_len != DWC_strlen(name))) )
    {
        return DWC_BADPARAMS;
    }
    /* name_len include null-terminator */
    if( name_len > DW_MAX_ADDR_LEN ) 
    {
        return DWC_TOOLONG;
    }
    switch( side )
    {
        case DWC_IPC_SIDE_LOCAL:
            fulladdr = &ipc->local;
            break;
        case DWC_IPC_SIDE_REMOTE:
            fulladdr = &ipc->remote;
            break;
        default:
            return DWC_BADPARAMS;
    }
    /* memset( fulladdr->name, 0, DW_MAX_ADDR_LEN ); */
    if( name )
    {                  
        strncpy( &fulladdr->name[0], name, DW_MAX_ADDR_LEN );
        fulladdr->name_len = DWC_strlen(fulladdr->name);
    }        
    else
    {
        fulladdr->name[0] = NULLCHAR;
        fulladdr->name_len = 0;
    }
    va_start( args, name_len );
    switch( ipc->type )
    {
        case DWC_IPC_TCP_SOCKET:                                   
            error = addr_tcp_op( cmd, &fulladdr->addr.sa_tcp, fulladdr->name, args );
            break;
#if defined(__UNIX_DW)
        case DWC_IPC_UNIX_SOCKET:
            error = addr_unix_op( cmd, &fulladdr->addr.sa_unix, fulladdr->name );
            goto error_exit;
#endif
        default:
            error = DWC_BADIPC;
            goto error_exit;    
    }
    if( error != DWC_NOERROR )
    {
        goto error_exit;
    }
    error = DWC_NOERROR;    
  error_exit:
    va_end( args );
    return error;
}

/* -------------------------------------------------- */
DWC_error_code_t bind_ipc ( DWC_ipc_t ipc )
{
    DWC_error_code_t  error;
    
    PARAM_ASSERT( ipc );
    
    if( ipc->binded )
    {
        return DWC_ALREADYPERFORM;
    }
    switch( ipc->type )
    {
        case DWC_IPC_TCP_SOCKET:
            error = bind_tcp_socket( ipc->handle.hnd_tcp, &ipc->local.addr.sa_tcp );           
            break;
#if defined(__UNIX_DW)
        case DWC_IPC_UNIX_SOCKET:
            error = bind_unix_socket( ipc->handle.hnd_unix, &ipc->local.addr.sa_unix );
            break;
#endif
        default:
            return DWC_BADIPC;
    }
    if( error == DWC_NOERROR )
    {
        ipc->binded = TRUE;
    }
    return error;
}

/* -------------------------------------------------- */
DWC_error_code_t listen_ipc ( DWC_ipc_t ipc, int backlog )
{
    DWC_error_code_t  error;
    
    PARAM_ASSERT( ipc );
    PARAM_ASSERT( (0 < backlog) && (backlog < MAX_THREADS) );
    
    if( ipc->listed || ipc->connected )
    {
        return DWC_ALREADYPERFORM;
    }
    switch( ipc->type )
    {
        case DWC_IPC_TCP_SOCKET:
            error = listen_tcp_socket( ipc->handle.hnd_tcp, backlog );           
            break;
#if defined(__UNIX_DW)
        case DWC_IPC_UNIX_SOCKET:
            error = listen_unix_socket( ipc->handle.hnd_unix, backlog );
            break;
#endif
        default:
            return DWC_BADIPC;
    }
    if( error == DWC_NOERROR )
    {
        ipc->listed = TRUE;
    }
    return error;
}

/* -------------------------------------------------- */
DWC_error_code_t accept_ipc ( DWC_ipc_t ipc, DWC_ipc_t* ipc_ptr, DWC_timeout_t timeout )
{
    DWC_error_code_t  error;
    
    PARAM_ASSERT( ipc );
    PARAM_ASSERT( ipc_ptr );
    
    if( !ipc->listed )
    {
        return DWC_NOTREADY;
    }    
    error = create_ipc( ipc->type, ipc_ptr );
    if( error != DWC_NOERROR )
    {
        return error;
    }
    switch( ipc->type )
    {
        case DWC_IPC_TCP_SOCKET:
            error = accept_tcp_socket( ipc->handle.hnd_tcp, &(*ipc_ptr)->remote.addr.sa_tcp, &(*ipc_ptr)->handle.hnd_tcp );           
            if( error != DWC_NOERROR )
            {
                goto error_exit;
            }
            (*ipc_ptr)->remote.name[0] = NULLCHAR;
            (*ipc_ptr)->remote.name_len = 0;
            break;
#if defined(__UNIX_DW)
        case DWC_IPC_UNIX_SOCKET:
            error = accept_unix_socket( ipc->handle.hnd_unix, &(*ipc_ptr)->remote.addr.sa_unix, &(*ipc_ptr)->handle.hnd_unix );           
            if( error != DWC_NOERROR )
            {
                goto error_exit;
            }
            goto error_exit;
#endif
        default:
            error = DWC_BADIPC;
            goto error_exit;
    }
    memcpy( &(*ipc_ptr)->local, &ipc->local, sizeof(DWC_ipc_addr_t) );
    (*ipc_ptr)->connected = TRUE;
    return DWC_NOERROR;
    
  error_exit:
    delete_ipc( ipc_ptr );
    return error;
}

/* -------------------------------------------------- */
DWC_error_code_t connect_ipc ( DWC_ipc_t ipc, DWC_timeout_t timeout )
{
    DWC_error_code_t  error = DWC_COMMONERROR;
    
    PARAM_ASSERT( ipc );    
    
    if( ipc->listed )
    {
        return DWC_NOTREADY;
    }    
    switch( ipc->type )
    {
        case DWC_IPC_TCP_SOCKET:
            error = connect_tcp_socket( ipc->handle.hnd_tcp, &ipc->remote.addr.sa_tcp, timeout );           
            if( error == DWC_TIMED )
            {
                dprintLib("Expired timeout on IPC operation.\n");
                ipcState.errored = TRUE;
                ipcState.error = EIO;
            }
            break;
#if defined(__UNIX_DW)
        case DWC_IPC_UNIX_SOCKET:
            error = connect_unix_socket( ipc->handle.hnd_unix, &ipc->remote.addr.sa_unix, timeout );           
            if( error == DWC_TIMED )
            {
                dprintLib("Expired timeout on IPC operation.\n");
                ipcState.errored = TRUE;
                ipcState.error = EIO;
            }
            break;
#endif /* __UNIX_DW */
        default:
            error = DWC_BADIPC;
            break;
    }
    if( error == DWC_NOERROR )
    {
        ipc->connected = TRUE;
    }
    return error;
}

/* -------------------------------------------------- */
DWC_error_code_t send_ipc ( DWC_ipc_t ipc, DWC_pointer_t buf, DWC_length_t* len_ptr, DWC_timeout_t timeout )
{
    DWC_error_code_t    error = DWC_NOERROR;
    DWC_length_t        needbytes = (*len_ptr);
    DWC_length_t        sendbytes = 0;
    DWC_timer_hnd_t     tid;
    int                 how;
    
    PARAM_ASSERT( ipc );
    PARAM_ASSERT( buf );
    PARAM_ASSERT( len_ptr );
    PARAM_ASSERT( *len_ptr >= 0 );
    
    if( (*len_ptr) == 0 ) 
    {
        return DWC_NOERROR;
    }
    if( ipc->nonblocking )
    {
        while( sendbytes < needbytes  )
        {
            error = ipc->send_func( ipc->handle, (DWC_pointer_t)((char*)buf + sendbytes), len_ptr, 0 );
            if( error != DWC_NOERROR )
            {            
                dprintLib( "Error on waiting for nonblocking IPC.\n" );
                break;
            }
            sendbytes += (*len_ptr);
            if( sendbytes >= needbytes )
            {
                break;            
            }
            how = DWC_IPC_WAIT_SEND;
            error = ipc->wait_func( ipc->handle, &how, timeout );
            if( (error != DWC_NOERROR) || !(how & DWC_IPC_WAIT_SEND) )
            {
                if( error == DWC_TIMED )
                {
                    dprintLib( "Timeout for nonblocking IPC expired.\n" );
                }
                else
                {
                    dprintLib( "Error on waiting for nonblocking IPC.\n" );
                }
                break;
            }
            *len_ptr = needbytes - sendbytes;
        }
        *len_ptr = sendbytes;
    }
    else
    {
        tid = SetTimeout( timeout, ipc->cancel_call, &ipc->handle, FALSE );
        if( tid == INVALID_TIMER )
        {
            dprintLib( "Can't set timeout for IPC operations.\n" );
            ipcState.errored = TRUE;
            ipcState.error = EBUSY;
            return DWC_TIMED;
        }
		while( sendbytes < needbytes )
		{     
        	error = ipc->send_func( ipc->handle, (DWC_pointer_t)((char*)buf + sendbytes), len_ptr, 0 );
        	if( error != DWC_NOERROR )
        	{
            	if( ExpiredTimeout( tid ) )
            	{   
                	dprintLib("Expired timeout on IPC operation.\n");
                	ipcState.errored = TRUE;
                	ipcState.error = ETIMEDOUT;
                	error = DWC_TIMED;
            	}
            	else if( error != DWC_NOERROR )
            	{
                	dprintLib("Error ocurred during IPC operation:", ipc->error_func() );
            	}
				break;
			}
            sendbytes += (*len_ptr);
            *len_ptr = needbytes - sendbytes;
        }
        ClearTimeout( tid );
		*len_ptr = sendbytes;
    }
    return error;    
}

/* -------------------------------------------------- */
DWC_error_code_t recv_ipc ( DWC_ipc_t ipc, DWC_pointer_t buf, DWC_length_t* len_ptr, DWC_timeout_t timeout )
{
    DWC_error_code_t    error = DWC_NOERROR;
    DWC_length_t        needbytes = (*len_ptr);
    DWC_length_t        recvbytes = 0;
    DWC_timer_hnd_t     tid;
    int                 how;

    PARAM_ASSERT( ipc );
    PARAM_ASSERT( buf );
    PARAM_ASSERT( len_ptr );
    PARAM_ASSERT( *len_ptr >= 0 );
    
    if( (*len_ptr) == 0 ) 
    {
        return DWC_NOERROR;
    }
    if( ipc->nonblocking )
    {
        while( recvbytes < needbytes  )
        {
            error = ipc->recv_func( ipc->handle, (DWC_pointer_t)((char*)buf + recvbytes), len_ptr, 0 );
            if( error != DWC_NOERROR )
            {            
                dprintLib( "Error on recv for nonblocking IPC.\n" );
                break;
            }
            recvbytes += (*len_ptr);
            if( recvbytes >= needbytes )
            {
                break;            
            }
            how = DWC_IPC_WAIT_RECV;
            error = ipc->wait_func( ipc->handle, &how, timeout );
            if( (error != DWC_NOERROR) || !(how & DWC_IPC_WAIT_RECV) )
            {
                if( error == DWC_TIMED )
                {
                    dprintLib( "Timeout for nonblocking IPC expired.\n" );
                }
                else
                {
                    dprintLib( "Error on waiting for nonblocking IPC.\n" );
                }
                break;
            }
            *len_ptr = needbytes - recvbytes;
        }
        *len_ptr = recvbytes;
    }
    else
    {
        tid = SetTimeout( timeout, ipc->cancel_call, &ipc->handle, FALSE );
        if( tid == INVALID_TIMER )
        {
            dprintLib( "Can't set timeout for IPC operations.\n" );
            ipcState.errored = TRUE;
            ipcState.error = EBUSY;
            return DWC_TIMED;
        }    
#if defined(__WIN32_DW)
#define MSG_WAITALL   0
#endif
        error = ipc->recv_func( ipc->handle, buf, len_ptr, MSG_WAITALL );
        if( error != DWC_NOERROR )
        {
            if( ExpiredTimeout( tid ) )
            {   
                dprintLib("Expired timeout on IPC operation.\n");
                ipcState.errored = TRUE;
                ipcState.error = ETIMEDOUT;
                error = DWC_TIMED;
            }
            else if( error != DWC_NOERROR )
            {
                dprintLib("Error ocurred during IPC operation:", ipc->error_func() );
            }
        }
        ClearTimeout( tid );
    }
    return error;    
}

/* -------------------------------------------------- */
DWC_string_t last_ipc_error ( DWC_ipc_t ipc )
{
    PARAM_ASSERT( ipc );
    if( ipcState.errored )
    {
        ipcState.errored = FALSE;
        return strerror(ipcState.error);
    }
    return ipc->error_func();
}

#endif /* _DWC_IPC_C */
