#ifndef _DWC_LIB_C
#define _DWC_LIB_C

#ifndef lint
static const char rcsid[] = "$Id: dwc_lib.c,v 1.15 2001/06/14 06:06:34 cvser Exp $";
#endif

#include <stdio.h>
#include <stddef.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#if defined(__WIN32_DW)
#include <winsock.h> /* for ntoh?() */
#include <io.h>
#else /* __WIN32_DW */
#include <unistd.h>
#include <sys/stat.h>
#include <sys/param.h> /* for PATH_MAX */
#include <netinet/in.h> /* for ntoh?() */
#endif /* __WIN32_DW */

#include "dwc_lib.h"
#include "dwc_type.h"
#include "dwc_ipc.h"
#include "dwc_err.h"
#include "dwc_macro.h"
#include "dwc_log.h"

typedef struct {
    DWC_bool_t             inited;
    DWC_ipc_type_t         ipc_type;
    DWC_timeout_t          timeout;
    DWC_string_t           addr;
    DWC_length_t           addr_len; /* include null-terminator */
    unsigned short         port;
} DWC_lib_settings_t;

static DWC_msg_entry_t  DaemonMsgTable[] = {
    { DERR_NOERROR,            "ok"},
    { DERR_READ_ERR,           "read errors" },
    { DERR_WRITE_ERR,          "write errors" },
    { DERR_NOMEMORY,           "memory errors" },
    { DERR_CRC_ERROR,          "CRC errors" },
    { DERR_READSOCKET,         "socket errors" },
    { DERR_KNOWN_VIRUS,        "known viruses" },
    { DERR_UNKNOWN_VIRUS,      "unknown viruses" },
    { DERR_VIRUS_MODIFICATION, "virus modifications" },
    { DERR_FILE_IS_CURED,      "cured files" },
    { DERR_TIMEOUT,            "timeouts" },
    { DERR_SYMLINK,            "symlinks" },
    { DERR_NO_REGFILE,         "no regular files" },
    { DERR_SKIPPED,            "skipped files (bad format)" },
    { DERR_TOO_BIG,            "too big files" },
    { DERR_TOO_COMPRESSED,     "too high compressed ratio" },
    { DERR_BAD_CALL,           "invalid calls" },
    END_OF_TABLE
};
static const int  DaemonMsgTableSize = (sizeof(DaemonMsgTable) / sizeof(DWC_msg_entry_t)) - 1;

static DWC_msg_entry_t  LibMsgTable[] = {
     { DWL_NOERROR,       "no errors" },
     { DWL_COMMONERROR,   "general fault" },
     { DWL_IPCSYSTEM,     "internal transport subsystem" },
     { DWL_BADFILE,       "bad file passed to lib" },     
     { DWL_BADPARAMS,     "incorrect values of parameters" },
     { DWL_NOTSUPPORTED,  "not supported" },
     { DWL_ALREADYEXEC,   "try make operation that can be performed only once" },
     { DWL_RECVERROR,     "recv from daemon" },
     { DWL_SENDERROR,     "send to daemon" },
     { DWL_NOMEMORY,      "no memory for internal use" },
     { DWL_NOTINIT,       "library was not inited" },
     END_OF_TABLE
};
static const int  LibMsgTableSize = (sizeof(LibMsgTable) / sizeof(DWC_msg_entry_t)) - 1;

#if defined(__WIN32_DW)
#define open_file( fd, mode )       _open( (fd), (mode) | O_BINARY )
#define read_file( fd, buf, len )   _read( (fd), (buf), (len) )
#define lseek_file( fd, off, w )    _lseek( (fd), (off), (w) )
#define close_file( fd )            _close( (fd) )
#else /* __WIN32_DW */
#define open_file( fd, mode )       open( (fd), (mode) )
#define read_file( fd, buf, len )   read( (fd), (buf), (len) )
#define lseek_file( fd, off, w )    lseek( (fd), (off), (w) )
#define close_file( fd )            close( (fd) )
#endif /* __WIN32_DW */

static DWC_lib_settings_t   Lib;

static DWC_ipc_t            session_open( void );
static int                  session_close( DWC_ipc_t ipc );
static int                  showResult( DWC_ipc_t ipc, int ScanOptions, DWC_log_t* log );
static int                  readPidFile( const char* FileName );
static int                  getFileSize( int fd );
static int                  scanNormalFile( DWC_string_t FileName, DWC_length_t NameLen, int ScanOptions, int* result_ptr, DWC_log_t* log );
#if defined(__UNIX_DW)
static int                  scanDiskFile( DWC_string_t PathName, DWC_length_t PathLen, int ScanOptions, int* result_ptr, DWC_log_t* log );
#endif /* __UNIX_DW */
static int                  writeInt( DWC_ipc_t ipc, int n );
static int                  writeBuf( DWC_ipc_t ipc, DWC_pointer_t buf, DWC_length_t size );
static int                  readInt( DWC_ipc_t ipc, int* n_ptr );
static int                  readOutput( DWC_ipc_t ipc, DWC_pointer_t buf, DWC_length_t size );

/* ----------------------------------------------------------------- */
void outputLib( int Quiet, int Debug, int Syslog )
{
    if( Quiet == 0 )
        VideoLogEnabled = TRUE;
    else
        VideoLogEnabled = FALSE;
    if( Debug != 0 )
        DebugLogEnabled = TRUE;
    else
        DebugLogEnabled = FALSE;
    if( Syslog != 0 )
        LogSyslog = TRUE;
    else
        LogSyslog = FALSE;
}

/* ----------------------------------------------------------------------- */
void dprintLib( char const* fmt, ... )
{
    va_list  args;
    
    if( DebugLogEnabled && VideoLogEnabled )
    {
        va_start( args, fmt );
        InternalPrint( LOG_DEBUG, fmt, args );
        va_end( args );
    }
}

/* -------------------------------------------------- */
void wprintLib( char const* fmt, ... )
{
    va_list args;

    if( VideoLogEnabled )
    {
        va_start( args, fmt );
        InternalPrint( LOG_NOTICE, fmt, args );
        va_end( args );
    }
}

/* ----------------------------------------------------------------- */
static DWC_ipc_t session_open( void )
{
    DWC_ipc_t ipc = INVALID_IPC;
    DWC_error_code_t error;
    
    error = create_ipc( Lib.ipc_type, &ipc );
    if( error != DWC_NOERROR )
    {
        dprintLib("Create ipc channel failed.\n");
        return INVALID_IPC;
    }
    switch( Lib.ipc_type )
    {
        case DWC_IPC_TCP_SOCKET:
            error = addr_ipc_op( ipc, DWC_IPC_CMD_RESOLVE, DWC_IPC_SIDE_REMOTE, Lib.addr, Lib.addr_len );
            if( error != DWC_NOERROR )
            {
                dprintLib("Can't resolve remote address(%s):%s.\n", Lib.addr, last_ipc_error(ipc) );
                goto error_exit;
            }
            error = addr_ipc_op( ipc, DWC_IPC_CMD_SETPORT, DWC_IPC_SIDE_REMOTE, NULL, 0, Lib.port );
            if( error != DWC_NOERROR )
            {
                dprintLib("Can't set remote port.\n");
                goto error_exit;
            }
            break;
        case DWC_IPC_UNIX_SOCKET:
            error = addr_ipc_op( ipc, DWC_IPC_CMD_SETADDR, DWC_IPC_SIDE_REMOTE, Lib.addr, Lib.addr_len );
            if( error != DWC_NOERROR )
            {
                dprintLib("Can't set remote port.\n");
                goto error_exit;
            }
            break;
        default:
            goto error_exit;
    }
    error = connect_ipc( ipc, Lib.timeout );
    if( error != DWC_NOERROR )
    {
        wprintLib(last_ipc_error( ipc ));
        wprintLib("\n");
        goto error_exit;
    }
    return ipc;
    
  error_exit:
    delete_ipc( &ipc );
    return INVALID_IPC;
}

//--------------------------------------------------------------------------
static int session_close( DWC_ipc_t ipc )
{
    if( delete_ipc( &ipc ) != DWC_NOERROR )
    {
        dprintLib("Close ipc channel failed.\n");
        return DWL_IPCSYSTEM;
    }
    return DWL_NOERROR;
}

//--------------------------------------------------------------------------
int startupLib( int Type, int Timeout, ... )
{
    va_list         args;
    int             error;
    unsigned int    Port;
    char const*     addr_name = NULL;    
#if defined(__UNIX_DW)
    DWC_error_code_t ret;
#endif /* __UNIX_DW */
    
    if( Lib.inited )
        return DWL_ALREADYEXEC;
    memset( &Lib, 0, sizeof(Lib) );
    if( init_ipc_subsystem() != DWC_NOERROR )
    {
        dprintLib("Can't init internal ipc subsystem.\n");
        return DWL_IPCSYSTEM;
    }    
    if( Timeout >= 0 )
        Lib.timeout = Timeout;
    else
        Lib.timeout = DRWEBD_DEFAULT_TIMEOUT;
    va_start( args, Timeout );
    switch( Type )
    {
        case DWL_TYPE_TCP_SOCKET:
            addr_name = va_arg( args, const char* );
            if( addr_name && (addr_name[0] != NULLCHAR) )
            {
                Lib.addr = strdup( addr_name );
            }
            else
            {
                Lib.addr = strdup( DRWEBD_DEFAULT_IP );
            }
            if( !Lib.addr )
            {
                dprintLib("No memory for allocate address.\n");
                error = DWL_NOMEMORY;
                goto error_exit;
            }
            Lib.addr_len = DWC_strlen(Lib.addr);
            Port = va_arg( args, unsigned int );
            if( Port != 0 )
            {
                Lib.port = (unsigned short)(0xFFFF & Port);
            }            
            else
            {
                Lib.port = DRWEBD_DEFAULT_PORT;
            }
            Lib.ipc_type = DWC_IPC_TCP_SOCKET;
            dprintLib("Use Dr.Web daemon INET-address - <%s:%d>, timeout(%u sec %u mcsec)\n", Lib.addr, Lib.port, Lib.timeout/1000000, Lib.timeout%1000000 );
            break;
#if defined(__UNIX_DW)        
        case DWL_TYPE_UNIX_SOCKET:
            addr_name = va_arg( args, const char* );
            if( addr_name && (addr_name[0] != NULLCHAR) )
            {
                Lib.addr = strdup( addr_name );
            }
            else
            {
                Lib.addr = strdup( DRWEBD_DEFAULT_SOCKETFILE );
            }
            if( !Lib.addr )
            {
                dprintLib("No memory for allocate address.\n");
                error = DWL_NOMEMORY;
                goto error_exit;
            }
            Lib.addr_len = DWC_strlen( Lib.addr );
            Lib.port = 0;            
            Lib.ipc_type = DWC_IPC_UNIX_SOCKET;
            dprintLib("Use Dr.Web daemon UNIX-address - <%s>, timeout(%u sec %u mcsec)\n", Lib.addr, Lib.timeout/1000000, Lib.timeout%1000000 );
            break;
        case DWL_TYPE_READPIDFILE:
            addr_name = va_arg( args, const char* );
            ret = readPidFile( addr_name );
            if( ret != DWL_NOERROR  )
            {
                dprintLib("Error during proccess pidfile.\n");
                error = ret;
                goto error_exit;
            }
            break;
#endif /* __UNIX_DW */
        case DWL_TYPE_DEFAULT:
            Lib.ipc_type = DWC_IPC_TCP_SOCKET;
            Lib.addr = strdup( DRWEBD_DEFAULT_IP );
            if( !Lib.addr )
            {
                dprintLib("No memory for allocate address.\n");
                error = DWL_NOMEMORY;
                goto error_exit;
            }
            Lib.addr_len = DWC_strlen( Lib.addr );
            Lib.port = DRWEBD_DEFAULT_PORT;
            dprintLib("Use defaults library settings: tcp/%s:%u, timeout(%u sec %u mcsec)\n", Lib.addr, Lib.port, Lib.timeout/1000000, Lib.timeout%1000000 );
            break;
        default:
            dprintLib("Unknown transport type requested.\n");
            error = DWL_BADPARAMS;            
            goto error_exit;
    }
    va_end( args );
    Lib.inited = TRUE;
    return DWL_NOERROR;
    
  error_exit:
    va_end( args );
    return error;
}

//--------------------------------------------------------------------------
int cleanupLib( void )
{
    if( !Lib.inited )
        return DWL_ALREADYEXEC;
    Lib.inited = FALSE;
    if( Lib.addr )
    {
        DWC_Free( Lib.addr );
    }
    memset( &Lib, 0, sizeof(Lib) );    
    if( dinit_ipc_subsystem() != DWC_NOERROR )
    {
        dprintLib("De-init internal ipc subsystem failed.\n");
        return DWL_IPCSYSTEM;
    }
    return DWL_NOERROR;
}

//--------------------------------------------------------------------------
static int readPidFile( const char* FileName )
{
    int          pid = 0, port = 0, error = DWL_NOERROR;
    FILE        *pidfile = NULL;
    char         socket[256] = { 0 };
    char         buffer[256] = { 0 };
    char        *pbuffer = NULL;
    char        *ptok = NULL;
    const char  *filename = NULL;
  
    if( !FileName || (FileName[0] == NULLCHAR) )
    {
        filename = DRWEBD_DEFAULT_PIDFILE;
        dprintLib("Using default pidfile name: <%s>.\n", filename);
    }
    else
    {
        filename = FileName;
    }
    pidfile = fopen( filename, "rb" );
    if( !pidfile )
    {
        dprintLib( "Can`t open pidfile %s.\n", filename );
        return DWL_BADFILE;
    }
    pbuffer = fgets( buffer, sizeof(buffer)-2, pidfile );
    if( !pbuffer )
    {
        dprintLib( "Can`t read string with pid from pidfile.\n" );
        return DWL_BADFILE;
    }
    buffer[sizeof(buffer)-1] = NULLCHAR;
    if( sscanf( pbuffer, "%d", &pid ) != 1 )
    {
        dprintLib( "Bad format of pid in pidfile.\n" );
        return DWL_BADFILE;
    }
    pbuffer = fgets( buffer, sizeof(buffer)-2, pidfile );
    if( !pbuffer )
    {
        dprintLib( "Can`t read string with address from pidfile.\n" );
        return DWL_BADFILE;
    }
    buffer[sizeof(buffer)-1] = NULLCHAR;
    if( buffer[0] == '/' )
    {   /* unix domain address */
        if( sscanf( pbuffer, "%255s", socket ) != 1 )
        {
            dprintLib( "Can`t read string with unix-address from pidfile.\n" );
            return DWL_BADFILE;
        }  
        if( socket[0] == NULLCHAR )
        {
            dprintLib( "Bad format of unix-address in pidfile.\n" );
            return DWL_BADFILE;            
        }
        if( Lib.addr )
        {
            free( Lib.addr );
        }
        Lib.addr = strdup( socket );
        if( Lib.addr )
        {
            Lib.addr_len = DWC_strlen( Lib.addr );
            Lib.ipc_type = DWC_IPC_UNIX_SOCKET;
            dprintLib( "Probably daemon use UNIX socket <%s>.\n", Lib.addr );
        }
        else
        {
            error = DWL_NOMEMORY;
        }            
    }
    else
    {   /* inet domain address */
        ptok = strrchr( pbuffer, ':' );
        if( !ptok )
        {
            dprintLib( "Can not parse string with inet-address from pidfile.\n" );
            return DWL_BADFILE;                    
        }
        if( (ptok - pbuffer) > DWC_Min( sizeof(socket)-1, sizeof(buffer)-2 ) )
        {
            dprintLib( "String with host name too long in pidfile.\n" );
            return DWL_BADFILE;                    
        }
        *ptok = NULLCHAR;        
        ++ptok;
        if( sscanf( ptok, "%d", &port ) != 1 )
        {
            dprintLib( "Can`t extract port number from pidfile.\n" );
            return DWL_BADFILE;                    
        }
        strncpy( socket, pbuffer, sizeof(socket)-1 );
        if( (port < 0) || (socket[0] == NULLCHAR) )
        {
            dprintLib( "Bad format of inet-address in pidfile.\n" );
            return DWL_BADFILE;            
        }
        if( Lib.addr )
        {
            free( Lib.addr );
        }
        Lib.addr = strdup( socket );
        if( Lib.addr )
        {
            Lib.addr_len = DWC_strlen( Lib.addr );
            Lib.port = port;
            Lib.ipc_type = DWC_IPC_TCP_SOCKET;
            dprintLib( "Probably daemon INET address <%s:%d>.\n", Lib.addr, Lib.port );            
        }
        else
        {
            error = DWL_NOMEMORY;
        }    
    }
    fclose( pidfile );
    return error;  
}

//--------------------------------------------------------------------------
int showVersion( char* log, size_t* loglen )
{
    int ver = 0;
    DWC_ipc_t ipc;
    DWC_log_t logbuf;
    DWC_log_t *plog;

    if( !Lib.inited )
    {
        dprintLib("Library wasn't startuped.\n");
        return DWL_NOTINIT;
    }    
    ipc = session_open();
    if( ipc == INVALID_IPC )
    {
        return DWL_IPCSYSTEM;
    }
    if( writeInt( ipc, DRWEBD_VERSION_CMD ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( readInt( ipc, &ver ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( log && loglen )
    {
        logbuf.line = log;
        logbuf.cur_len = 0;
        logbuf.max_len = *loglen;        
        plog = &logbuf;
    }
    else
    {
        plog = NULL;
    }
    StartLog( plog );
    AddLog( plog, "Dr.Web daemon version: %u.%u\n", ver/100, ver%100 );
    EndLog( plog );
    if( plog )
    {
        *loglen = plog->cur_len;
    }    
    return session_close( ipc );
    
  error_exit:
    session_close( ipc );
    return DWL_IPCSYSTEM;
}

//--------------------------------------------------------------------------
int showID( char* log, size_t* loglen )
{
    char buf[ 1024 ];
    int id_len = 0;
    DWC_ipc_t ipc;
    DWC_log_t logbuf;
    DWC_log_t *plog;

    if( !Lib.inited )
    {
        dprintLib("Library wasn't startuped.\n");
        return DWL_NOTINIT;
    }
    ipc = session_open();
    if( ipc == INVALID_IPC )
    {
        return DWL_IPCSYSTEM;
    }
    if( writeInt( ipc, DRWEBD_IDSTRING_CMD ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( readInt( ipc, &id_len ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( id_len > sizeof(buf)-1 )
        id_len = sizeof(buf)-1;
    if( readOutput( ipc, buf, id_len ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    buf[ sizeof(buf)-1 ] = '\0';
    if( log && loglen )
    {
        logbuf.line = log;
        logbuf.cur_len = 0;
        logbuf.max_len = *loglen;        
        plog = &logbuf;
    }
    else
    {
        plog = NULL;
    }
    StartLog( plog );
    AddLog( plog, buf );
    AddLog( plog, "\n" );
    EndLog( plog );
    if( plog )
    {
        *loglen = plog->cur_len;
    }
    return session_close( ipc );
    
  error_exit:
    session_close( ipc );
    return DWL_IPCSYSTEM;
}

//--------------------------------------------------------------------------
int showBases( char* log, size_t* loglen )
{
    int i;
    int bases_num = 0;
    int name_len = 0;
    int vir_num = 0;
    char base_filename[ 1024 ];
    DWC_ipc_t ipc;
    DWC_log_t logbuf;
    DWC_log_t *plog;

    if( !Lib.inited )
    {
        dprintLib("Library wasn't startuped.\n");
        return DWL_NOTINIT;
    }
    ipc = session_open();
    if( ipc == INVALID_IPC )
    {
        return DWL_IPCSYSTEM;
    }
    if( writeInt( ipc, DRWEBD_BASEINFO_CMD ) != DWL_NOERROR )
    {
        goto error_exit;        
    }
    if( readInt( ipc, &bases_num ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( log && loglen )
    {
        logbuf.line = log;
        logbuf.cur_len = 0;
        logbuf.max_len = *loglen;        
        plog = &logbuf;
    }
    else
    {
        plog = NULL;
    }
    StartLog( plog );
    for( i = 0; i < bases_num; ++i )
    {
        if( readInt( ipc, &name_len ) != DWL_NOERROR )
        {
            goto error_exit;
        }
        if( name_len > sizeof(base_filename)-1 )
        {
            name_len = sizeof(base_filename)-1;
        }
        if( readOutput( ipc, base_filename, name_len ) != DWL_NOERROR )
        {
            goto error_exit;
        }
        base_filename[ sizeof(base_filename)-1 ] = '\0';
        if( readInt( ipc, &vir_num ) != DWL_NOERROR )
        {
            goto error_exit;
        }
        AddLog( plog, "Virus base: %s (%u records)\n", base_filename, vir_num );
    }
    EndLog( plog );
    if( plog )
    {
        *loglen = plog->cur_len;
    }
    return session_close( ipc );
    
  error_exit:
    session_close( ipc );
    return DWL_IPCSYSTEM;
}

//--------------------------------------------------------------------------
int scanBuffer( void* Buffer, size_t BufferLen, int ScanOptions, int* result_ptr, char* log, size_t* loglen )
{
    DWC_ipc_t ipc;
    DWC_log_t logbuf;
    DWC_log_t *plog;
    
    if( !Lib.inited )
    {
        dprintLib("Library wasn't startuped.\n");
        return DWL_NOTINIT;
    }
    if( !result_ptr )
    {
        return DWL_BADPARAMS;
    }
    ipc = session_open();
    if( ipc == INVALID_IPC )
    {
        return DWL_IPCSYSTEM;
    }
    if( writeInt( ipc, DRWEBD_SCAN_CMD ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeInt( ipc, ScanOptions ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeInt( ipc, 0 ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeInt( ipc, BufferLen ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeBuf( ipc, Buffer, BufferLen ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( readInt( ipc, result_ptr )  )
    {
        goto error_exit;
    }
    if( log && loglen )
    {
        logbuf.line = log;
        logbuf.cur_len = 0;
        logbuf.max_len = *loglen;        
        plog = &logbuf;
    }
    else
    {
        plog = NULL;
    }
    if( showResult( ipc, ScanOptions, plog ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( plog )
    {
        *loglen = plog->cur_len;
    }
    return session_close( ipc );
    
  error_exit:
    session_close( ipc );
    return DWL_IPCSYSTEM;
}

//--------------------------------------------------------------------------
int scanFile( char* Path, size_t PathLen, int FileType, int ScanOptions, int* result_ptr, char* log, size_t* loglen )
{
    int       retcode;
    DWC_log_t logbuf;    
    DWC_log_t *plog;
    
    if( !Lib.inited )
    {
        dprintLib("Library wasn't startuped.\n");
        return DWL_NOTINIT;
    }
    if(    (Path && (PathLen == 0))
		|| (Path && (PathLen > PATH_MAX))
		|| (!Path && (PathLen != 0))
        || !result_ptr )
    {
        return DWL_BADPARAMS;
    }
    if( log && loglen )
    {
        logbuf.line = log;
        logbuf.cur_len = 0;
        logbuf.max_len = *loglen;        
        plog = &logbuf;
    }
    else
    {
        plog = NULL;
    }
    switch( FileType )
    {
        case DWL_FT_NORMALFILE:
            retcode = scanNormalFile( Path, PathLen, ScanOptions, result_ptr, plog );
            break;
#if defined(__UNIX_DW)
        case DWL_FT_DISKFILE:
            retcode = scanDiskFile( Path, PathLen, ScanOptions, result_ptr, plog );
            break;
#endif
        default:
            dprintLib("Unsupported file type.\n");
            retcode = DWL_NOTSUPPORTED;
            break;
    }    
    if( plog )
    {
        *loglen = plog->cur_len;
    }
    return retcode;
}

//--------------------------------------------------------------------------
static int getFileSize( int fd )
{
    off_t fs;
 
    fs = lseek_file( fd, 0, SEEK_END );
    if( fs < 0 )
    {
        return -1;
    }
    if( lseek_file( fd, 0, SEEK_SET ) < 0 )
    {
        return -1;
    }
    return (int)fs;
}

//--------------------------------------------------------------------------
static int scanNormalFile( DWC_string_t FileName, DWC_length_t NameLen, int ScanOptions, int* result_ptr, DWC_log_t* log )
{
    DWC_ipc_t ipc;
    char buf[ 0x1000 ];
    int towrite = sizeof(buf);
    int fd;
    size_t fileSize;
    
    fd = open_file( FileName, O_RDONLY );
    if( fd < 0 )
    {
        wprintLib("Can't open file.\n");
        return DWL_BADFILE;
    }
    fileSize = getFileSize( fd );
    if( fileSize < 0 )
    {
        wprintLib("Can't determine file size.\n");
    }    
    ipc = session_open();
    if( ipc == INVALID_IPC )
    {
        return DWL_IPCSYSTEM;
    }
    if( writeInt( ipc, DRWEBD_SCAN_CMD ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeInt( ipc, ScanOptions ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeInt( ipc, NameLen ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeBuf( ipc, FileName, NameLen ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeInt( ipc, fileSize ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    while( fileSize > 0 )
    {
        if( towrite > fileSize )
        {
            towrite = fileSize;
        }
        if( read_file( fd, buf, towrite ) != towrite )
        {
            goto error_exit_2;
        }
        if( writeBuf( ipc, buf, towrite ) != DWL_NOERROR )
        {
            goto error_exit;
        }
        fileSize -= towrite;
    }
    if( readInt( ipc, result_ptr ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( showResult( ipc, ScanOptions, log ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    close_file( fd );
    return session_close( ipc );
    
  error_exit:
    close_file( fd );
    session_close( ipc );
    return DWL_IPCSYSTEM;
    
  error_exit_2:
    close_file( fd );
    session_close( ipc );
    return DWL_BADFILE;
}

//--------------------------------------------------------------------------
#if defined(__UNIX_DW)
static int scanDiskFile( DWC_string_t PathName, DWC_length_t PathLen, int ScanOptions, int* result_ptr, DWC_log_t* log )
{
    DWC_ipc_t ipc;
    
    ipc = session_open();
    if( ipc == INVALID_IPC )
    {
        return DWL_IPCSYSTEM;
    }
    if( writeInt( ipc, DRWEBD_SCAN_CMD ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeInt( ipc, ScanOptions ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeInt( ipc, PathLen ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeBuf( ipc, PathName, PathLen ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeInt( ipc, 0 ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( readInt( ipc, result_ptr ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( showResult( ipc, ScanOptions, log ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    return session_close( ipc );
    
  error_exit:
    session_close( ipc );
    return DWL_IPCSYSTEM;
}
#endif /* __UNIX_DW */

//--------------------------------------------------------------------------
int scanFilePart( char* FileName, size_t NameLen, size_t Offset, size_t Length, int ScanOptions, int* result_ptr, char* log, size_t* loglen  )
{    
#if !defined(__UNIX_DW)
    return DWL_NOTSUPPORTED;
#else /* __UNIX_DW */
    DWC_ipc_t ipc;
    DWC_log_t logbuf;
    DWC_log_t *plog;

    if( !Lib.inited )
    {
        dprintLib("Library wasn't startuped.\n");
        return DWL_NOTINIT;
    }
    if( !result_ptr
		|| (FileName && (NameLen > PATH_MAX))
		|| (!FileName && (NameLen != 0))
        || (FileName && (NameLen == 0)) )
    {
        return DWL_BADPARAMS;
    }
    ipc = session_open();
    if( ipc == INVALID_IPC )
    {
        return DWL_IPCSYSTEM;
    }
    if( writeInt( ipc, DRWEBD_SCANPART_CMD ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeInt( ipc, Offset ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeInt( ipc, Length ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeInt( ipc, ScanOptions ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeInt( ipc, NameLen ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( writeBuf( ipc, FileName, NameLen ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( readInt( ipc, result_ptr ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    if( log && loglen )
    {
        logbuf.line = log;
        logbuf.cur_len = 0;
        logbuf.max_len = *loglen;        
        plog = &logbuf;
    }
    else
    {
        plog = NULL;
    }
    if( showResult( ipc, ScanOptions, plog ) != DWL_NOERROR )
    {
        goto error_exit;
    }
    return session_close( ipc );
    
  error_exit:
    session_close( ipc );
    return DWL_IPCSYSTEM;
#endif /* __UNIX_DW */
}

//--------------------------------------------------------------------------
const char* getLibError( int code )
{
    return findWhat( code, LibMsgTable );
}

//--------------------------------------------------------------------------
const char* getDaemonCode( int code )
{
    return findWhat( code, DaemonMsgTable );
}

//--------------------------------------------------------------------------
static int showResult( DWC_ipc_t ipc, int ScanOptions, DWC_log_t* log )
{
    int i;
    int vir_num = 0;
    int virname_len = 0;
    char virname[ 1024 ];
    int line_num = 0;
    int line_len = 0;
    char logline[ 1024 ];
    int codes_num = 0;
    int what = 0;
    int val = 0;

    StartLog( log );
    if( (ScanOptions & DRWEBD_RETURN_VIRUSES) == DRWEBD_RETURN_VIRUSES )
    {
        if( readInt( ipc, &vir_num ) != DWL_NOERROR )
        {
            goto error_exit;
        }

        for( i = 0; i < vir_num; i++ )
        {
            if( readInt( ipc, &virname_len ) != DWL_NOERROR )
            {
                goto error_exit;
            }
            if( virname_len > sizeof(virname)-1 )
                virname_len = sizeof(virname);
            if( readOutput( ipc, virname, virname_len ) != DWL_NOERROR )
            {
                goto error_exit;
            }
            AddLog(log, virname);
            AddLog(log, "\n");
        }
    }

    if( (ScanOptions & DRWEBD_RETURN_REPORT) == DRWEBD_RETURN_REPORT )
    {
        if( readInt( ipc, &line_num ) != DWL_NOERROR )
        {
            goto error_exit;
        }
        
        for( i = 0; i < line_num; ++i )
        {
            if( readInt( ipc, &line_len ) != DWL_NOERROR )
            {
                goto error_exit;
            }
            if( line_len > sizeof(logline)-1 )
                line_len = sizeof(logline)-1;
            if( readOutput( ipc, logline, line_len ) != DWL_NOERROR )
            {
                goto error_exit;
            }
            AddLog(log, logline);
        }
    }

    if( (ScanOptions & DRWEBD_RETURN_CODES) == DRWEBD_RETURN_CODES )
    {
        if( readInt( ipc, &codes_num ) != DWL_NOERROR )
        {
            goto error_exit;
        }
        for( i = 0; i < codes_num; ++i )
        {
            if( (readInt( ipc, &what ) != DWL_NOERROR)
                ||
                (readInt( ipc, &val ) != DWL_NOERROR) )
            {
                goto error_exit;
            }
            AddLog( log, "%s : %d\n", findWhat( what, DaemonMsgTable ), val );
        }	
    }
    EndLog( log );
    return DWL_NOERROR;
    
  error_exit:
    EndLog( log );
    return DWL_IPCSYSTEM;
}

//--------------------------------------------------------------------------
static int writeInt( DWC_ipc_t ipc, int n )
{    
    PARAM_ASSERT( ipc );
    
    n = htonl( n );
    return writeBuf( ipc, &n, sizeof(n) );
}

//--------------------------------------------------------------------------
static int writeBuf( DWC_ipc_t ipc, DWC_pointer_t buf, DWC_length_t size )
{
    DWC_error_code_t error;
    
    PARAM_ASSERT( ipc );
    PARAM_ASSERT( buf );
    PARAM_ASSERT( size > 0 );
    
    error = send_ipc( ipc, buf, &size, Lib.timeout );
    if( error )
    {
        wprintLib("Send failed (%d - %s).\n", error, last_ipc_error(ipc));
        dprintLib("Sended %d bytes.\n", size);
        return DWL_SENDERROR;
    }
    return DWL_NOERROR;
}

//--------------------------------------------------------------------------
static int readInt( DWC_ipc_t ipc, int* n_ptr )
{
    int error;

    PARAM_ASSERT( ipc );
    PARAM_ASSERT( n_ptr );
    
    error = readOutput( ipc, n_ptr, sizeof(int) );
    if( error == DWL_NOERROR )
    {
        *n_ptr = ntohl((*n_ptr));
    }
    return error;
}

//--------------------------------------------------------------------------
static int readOutput( DWC_ipc_t ipc, DWC_pointer_t buf, DWC_length_t size )
{
    DWC_error_code_t error;

    PARAM_ASSERT( ipc );
    PARAM_ASSERT( buf );
    PARAM_ASSERT( size > 0 );
    
    error = recv_ipc( ipc, buf, &size, Lib.timeout );
    if( error != DWC_NOERROR )
    {
        wprintLib("Recv failed (%s).\n", last_ipc_error(ipc));
        dprintLib("Received %d bytes.\n", size);
        return DWL_RECVERROR;
    }
    return DWL_NOERROR;
}

//--------------------------------------------------------------------------
#endif
